@@ -686,6 +686,16 @@ weak_table_put_x (scm_t_weak_table *table, unsigned long hash,
         }
     }
           
+  /* Fast path for updated values for existing entries of weak-key
+     tables.  */
+  if (table->kind == SCM_WEAK_TABLE_KIND_KEY &&
+      entries[k].hash == hash &&
+      entries[k].key == SCM_UNPACK (key))
+    {
+      entries[k].value = SCM_UNPACK (value);
+      return;
+    }
+
   if (entries[k].hash)
     unregister_disappearing_links (&entries[k], table->kind);
   else
@@ -912,9 +922,6 @@ assq_predicate (SCM x, SCM y, void *closure)
 SCM
 scm_weak_table_refq (SCM table, SCM key, SCM dflt)
 {
-  if (SCM_UNBNDP (dflt))
-    dflt = SCM_BOOL_F;
-  
   return scm_c_weak_table_ref (table, scm_ihashq (key, -1),
                                assq_predicate, SCM_UNPACK_POINTER (key),
                                dflt);
